import enum
import json
from typing import Optional, Any, Dict


class RpcErrorCode(enum.Enum):
    PARSE_ERROR = -32700
    INVALID_REQUEST = -32600
    METHOD_NOT_FOUND = -32601
    INVALID_PARAMS = -32602
    INTERNAL_ERROR = -32603

    def message(self) -> str:
        return message_map[self]


message_map = {
    RpcErrorCode.PARSE_ERROR: "Parse error",
    RpcErrorCode.INVALID_REQUEST: "Invalid request",
    RpcErrorCode.METHOD_NOT_FOUND: "Invalid method",
    RpcErrorCode.INVALID_PARAMS: "Invalid params",
    RpcErrorCode.INTERNAL_ERROR: "Internal error",
}


class RpcError(Exception):
    def __init__(
        self,
        code: RpcErrorCode,
        request_id: Optional[str],
        data: Optional[Any],
        *,
        message: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.code = code
        if message:
            self.message = message
        else:
            self.message = code.message()

        self.data = data
        self.id = request_id

    def __str__(self):
        return (
            f"RpcError<code: {self.code}, message: {self.message}, data:"
            f" {self.data}>"
        )

    def to_json(self) -> Dict[str, Any]:
        fields = {
            "code": self.code.value,
            "message": self.message,
        }
        if self.data is not None:
            fields["data"] = self.data
        return fields

    def to_jsons(self, encoder=json.dumps) -> str:
        return encoder(self.to_json())

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "RpcError":
        return cls(
            RpcErrorCode(payload["code"]),
            None,
            payload.get("data"),
            message=payload.get("message"),
        )

    @classmethod
    def from_jsons(cls, payload: str, decoder=json.loads) -> "RpcError":
        return cls.from_json(decoder(payload))


class RpcParseError(RpcError):
    def __init__(
        self, request_id: Optional[str] = None, data: Optional[Any] = None
    ) -> None:
        super().__init__(RpcErrorCode.PARSE_ERROR, request_id, data)


class RpcInvalidRequest(RpcError):
    def __init__(
        self, request_id: Optional[str], data: Optional[Any] = None
    ) -> None:
        super().__init__(RpcErrorCode.INVALID_REQUEST, request_id, data)


class RpcMethodNotFound(RpcError):
    def __init__(
        self, request_id: Optional[str], data: Optional[Any] = None
    ) -> None:
        super().__init__(RpcErrorCode.METHOD_NOT_FOUND, request_id, data)


class RpcInvalidParams(RpcError):
    def __init__(
        self, request_id: Optional[str], data: Optional[Any] = None
    ) -> None:
        super().__init__(RpcErrorCode.INVALID_PARAMS, request_id, data)


class RpcInternalError(RpcError):
    def __init__(
        self, request_id: Optional[str], data: Optional[Any] = None
    ) -> None:
        super().__init__(RpcErrorCode.INTERNAL_ERROR, request_id, data)



from .connection import (
    WsRpcConnection,
    WsRpcOpts,
    WsException,
    WsNotConnected,
    WsRpcTimedOut,
)

__all__ = [
    "WsRpcConnection",
    "WsRpcOpts",
    "WsException",
    "WsNotConnected",
    "WsRpcTimedOut",
]



import asyncio
from typing import Dict, Optional

from jsonrpc.types import JsonRpcResponse


class ResponseQueue:
    """
    Queue for matching RPC requests with responses.
    Usage of this class expects all RPC requests to contain request IDs.

    This class doesn't clean up its memory usage, so if the corresponding RPC
    server doesn't respond to request IDs this queue will continue growing.

    This class does not properly handle multiple requests with the same request_id,
    but you probably shouldn't be doing that anyway.
    """

    def __init__(self):
        self.message_notifiers: Dict[str, asyncio.Event] = {}
        self.message_by_request_id: Dict[str, JsonRpcResponse] = {}

    async def put(self, message: JsonRpcResponse) -> None:
        """
        Upon receiving an RPC response, `put` should be called to signal to the request waiter
        that a response is ready.

        :param message: RPC response
        """
        request_id = message.id
        if request_id is None:
            raise ValueError(
                "Response queue cannot accept RPC messages with no request ID."
            )
        self.message_by_request_id[request_id] = message

        if request_id in self.message_notifiers:
            self.message_notifiers[request_id].set()

    async def get(self, request_id: str) -> Optional[JsonRpcResponse]:
        """
        After dispatching an RPC request, `get` should be called to await its response.

        :param request_id: RPC request id
        :return: RPC response
        """
        if request_id == "":
            raise ValueError("request_id cannot be empty")

        # message has already been answered: return immediately
        if request_id in self.message_by_request_id:
            message = self.message_by_request_id[request_id]
            self._cleanup(request_id)
            return message

        if request_id in self.message_notifiers:
            event = self.message_notifiers[request_id]
        else:
            event = asyncio.Event()
            self.message_notifiers[request_id] = event

        await event.wait()

        if request_id not in self.message_by_request_id:
            return None

        message = self.message_by_request_id[request_id]
        self._cleanup(request_id)
        return message

    def _cleanup(self, request_id: str) -> None:
        self.message_notifiers.pop(request_id, None)
        self.message_by_request_id.pop(request_id, None)
